/*
  Hi-Lo All-03 & Sunshine Expro60 pin testing software.

  Compile with open watcom: wcl -l=dos all3test.c
*/

#define LICENCE_TEXT                                                    \
"To the extent possible under law, the author has dedicated all "       \
"copyright and related and neighboring rights to this software to "     \
"the public domain worldwide. This software is distributed without "    \
"any warranty. You should have received a copy of the CC0 Public "      \
"Domain Dedication along with this software. If not, see "              \
"http://creativecommons.org/publicdomain/zero/1.0/\n\n"                 \
"alexandre.becoulet@free.fr - 2024\n"

#include <conio.h>
#include <bios.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <i86.h>
#include <unistd.h>

#include <graph.h>

static char gui_query_bar(const char *message);

#define PIN_COUNT 40

/* pin modes */
#define PIN_M_TTLLOW 0
#define PIN_M_INPUT  1
#define PIN_M_VCC    2
#define PIN_M_VOP    3
#define PIN_M_VOP25  4  /* VOP - 2.5v */
#define PIN_M_VHH    5
#define PIN_M_VHH2   6  /* VHH high current */
#define PIN_M_CLK    7
#define PIN_M_100N   8  /* 100nF capacitor */
#define PIN_M_GND    9  /* GND one pin at a time via a 3to8 decoder */
#define PIN_M_GND2   10 /* GND via dedicated register bit */

#define PIN_M_COUNT  11

static const char *pin_cap_desc[PIN_M_COUNT] =
{
  "Low",  "Input",  "VCC",  "VOP",
  "VOP-2.5v",  "VHH",  "VHH Alt",
  "Clock",  "100nF",  "GND",  "GND Alt"
};

#define _B(x) (1 << (x))

static const uint16_t pin_capability[PIN_COUNT] = {
  /*  1 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VCC) | _B(PIN_M_VOP) | _B(PIN_M_GND),
  /*  2 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_CLK),
  /*  3 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_CLK),
  /*  4 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT),
  /*  5 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VCC) | _B(PIN_M_VOP),
  /*  6 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VOP),
  /*  7 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VCC) | _B(PIN_M_VOP) | _B(PIN_M_VOP25),
  /*  8 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT),
  /*  9 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VCC) | _B(PIN_M_VOP),
  /* 10 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP),
  /* 11 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP) | _B(PIN_M_GND),
  /* 12 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP),
  /* 13 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP),
  /* 14 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP),
  /* 15 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP),
  /* 16 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP),
  /* 17 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP),
  /* 18 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP) | _B(PIN_M_CLK),
  /* 19 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VCC) | _B(PIN_M_VOP) | _B(PIN_M_CLK),
  /* 20 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VCC) | _B(PIN_M_VOP) | _B(PIN_M_GND),
  /* 21 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VCC) | _B(PIN_M_VOP) | _B(PIN_M_GND),
  /* 22 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP),
  /* 23 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VOP),
  /* 24 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VCC) | _B(PIN_M_VOP),
  /* 25 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VCC) | _B(PIN_M_VOP) | _B(PIN_M_VOP25),
  /* 26 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VCC) | _B(PIN_M_VOP),
  /* 27 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VCC) | _B(PIN_M_VOP),
  /* 28 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VHH2) | _B(PIN_M_VCC) | _B(PIN_M_VOP) | _B(PIN_M_GND),
  /* 29 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VHH2) | _B(PIN_M_VCC) | _B(PIN_M_VOP) | _B(PIN_M_GND),
  /* 30 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VHH2) | _B(PIN_M_VCC) | _B(PIN_M_VOP) | _B(PIN_M_GND) | _B(PIN_M_GND2) | _B(PIN_M_100N),
  /* 31 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VHH2) | _B(PIN_M_VCC) | _B(PIN_M_VOP),
  /* 32 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VHH) | _B(PIN_M_VHH2) | _B(PIN_M_VCC) | _B(PIN_M_VOP) | _B(PIN_M_100N),
  /* 33 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT),
  /* 34 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VCC),
  /* 35 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT),
  /* 36 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VCC) | _B(PIN_M_VOP),
  /* 37 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT),
  /* 38 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT),
  /* 39 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT),
  /* 40 */ _B(PIN_M_TTLLOW) | _B(PIN_M_INPUT) | _B(PIN_M_VCC) | _B(PIN_M_GND)
};

/* register ids */
#define REG_TTLID0      0xe0   // TTL I/O for pins 1 to 8
#define REG_TTLID1      0xe1   // TTL I/O for pins 9 to 16
#define REG_TTLID2      0xe2   // TTL I/O for pins 17 to 24
#define REG_TTLID3      0xe3   // TTL I/O for pins 25 to 32
#define REG_TTLID4      0xe4   // TTL I/O for pins 33 to 40
#define REG_VOPID       0xe5   // VOP voltage
#define REG_VHHID       0xe6   // VHH voltage
#define REG_VCCID       0xe7   // VCC voltage
#define REG_VOPENID0    0xe8   // VOP enable for pins 1 to 8
#define REG_VOPENID1    0xe9   // VOP enable for pins 9 to 16
#define REG_VOPENID2    0xea   // VOP enable for pins 17 to 24
#define REG_VOPENID3    0xeb   // VOP enable for pins 25 to 32
#define REG_VOPENID4    0xec   // VOP enable for pins 33 to 40
#define REG_VCCENID0    0xed   // VCC enable for pins 40, 36, 34, 32, 31, 30, 29 and 28
#define REG_VCCENID1    0xee   // VCC enable for pins 27, 26, 9, 7, 5, 1, {25 and 24}
#define REG_OTHERENID0  0xef   // GND select for pins 20, 11, 1, 21, 28, 40, 30, and 29. + others.
#define REG_VHHENID0    0xf0   // VHH enable for pins 1 to 8, not implemented in hw
#define REG_VHHENID1    0xf1   // VHH enable for pins 9 to 16
#define REG_VHHENID2    0xf2   // VHH enable for pins 17 to 24
#define REG_VHHENID3    0xf3   // VHH enable for pins 25 to 32  (U23)
#define REG_VHHENID4    0xf4   // VHH enable for pins 33 to 40, not implemented in hw
#define REG_VHHENCID0   0xf5   // VHH enable for pins 28, 29, 30, 31 and 32  (U22)
#define REG_VHHENCID1   0xf6   // VHH enable, not implemented in hw
#define REG_OTHERENID1  0xf7   // VCC enable for pins 19, 20 and 21 + others.
#define REG_CODEID      0xf8   // device detection code register

static uint16_t base_addr = 0x2e0;

/* keep track of current device state */
static uint8_t pin_mode[PIN_COUNT];
static uint8_t reg_val[0x20];   /* value of registers starting at 0xe0 */

#define VCC_DAC_MIN     0
#define VCC_DAC_MAX  1020
#define VCC_DAC_STEP    4

#define VOP_DAC_LOW  1020
#define VOP_DAC_MIN     0
#define VOP_DAC_MAX  2550
#define VOP_DAC_STEP   10

#define VHH_DAC_LOW   510
#define VHH_DAC_MIN     0
#define VHH_DAC_MAX  1530
#define VHH_DAC_STEP    6

static int16_t vop_dac = VOP_DAC_LOW;
static int16_t vhh_dac = VHH_DAC_LOW;
static int16_t vcc_dac = VCC_DAC_MIN;

uint8_t vhh_count = 0;
uint8_t vop_count = 0;

static inline uint8_t all3_reg_in(uint8_t reg)
{
  outp(base_addr, reg);
  return inp(base_addr + 2);
}

static inline uint8_t all3_reg_get(uint8_t reg)
{
  return reg_val[reg - 0xe0];
}

static inline void all3_reg_set(uint8_t reg, uint8_t value)
{
  outp(base_addr, reg);
  outp(base_addr + 2, value);
  reg_val[reg - 0xe0] = value;
}

static inline void all3_reg_or(uint8_t reg, uint8_t value, uint8_t mask)
{
  all3_reg_set(reg, (reg_val[reg - 0xe0] & ~mask) | (value & mask));
}

static uint8_t all3_read_pin(uint16_t pin)
{
  uint8_t pin_reg = (pin - 1) / 8;
  uint8_t pin_bit = (pin - 1) % 8;
  return (all3_reg_in(REG_TTLID0 + pin_reg) >> pin_bit) & 1;
}

#define _R(reg, bit) (REG_##reg << 8 | /* entry valid */ 0x80 | bit)

/* Specifies register id and bit index to touch for most pin and mode combinations. */
static const uint16_t pin_reg_ops[PIN_COUNT][7] = {
  /*            INPUT         VCC                VOP              VOP25              VHH               VHH2  */
  /*  1 */ { 0, _R(TTLID0, 0), _R(VCCENID1, 5),   _R(VOPENID0, 0), 0,                 0,                0                },
  /*  2 */ { 0, _R(TTLID0, 1), 0,                 0,               0,                 0,                0                },
  /*  3 */ { 0, _R(TTLID0, 2), 0,                 0,               0,                 0,                0                },
  /*  4 */ { 0, _R(TTLID0, 3), 0,                 0,               0,                 0,                0                },
  /*  5 */ { 0, _R(TTLID0, 4), _R(VCCENID1, 4),   _R(VOPENID0, 4), 0,                 0,                0                },
  /*  6 */ { 0, _R(TTLID0, 5), 0,                 _R(VOPENID0, 5), 0,                 0,                0                },
  /*  7 */ { 0, _R(TTLID0, 6), _R(VCCENID1, 3),   _R(VOPENID0, 6), _R(OTHERENID0, 1), 0,                0                },
  /*  8 */ { 0, _R(TTLID0, 7), 0,                 0,               0,                 0,                0                },

  /*  9 */ { 0, _R(TTLID1, 0), _R(VCCENID1, 2),   _R(VOPENID1, 0), 0,                 _R(VHHENID1, 0),  0                },
  /* 10 */ { 0, _R(TTLID1, 1), 0,                 _R(VOPENID1, 1), 0,                 _R(VHHENID1, 1),  0                },
  /* 11 */ { 0, _R(TTLID1, 2), 0,                 _R(VOPENID1, 2), 0,                 _R(VHHENID1, 2),  0                },
  /* 12 */ { 0, _R(TTLID1, 3), 0,                 _R(VOPENID1, 3), 0,                 _R(VHHENID1, 3),  0                },
  /* 13 */ { 0, _R(TTLID1, 4), 0,                 _R(VOPENID1, 4), 0,                 _R(VHHENID1, 4),  0                },
  /* 14 */ { 0, _R(TTLID1, 5), 0,                 _R(VOPENID1, 5), 0,                 _R(VHHENID1, 5),  0                },
  /* 15 */ { 0, _R(TTLID1, 6), 0,                 _R(VOPENID1, 6), 0,                 _R(VHHENID1, 6),  0                },
  /* 16 */ { 0, _R(TTLID1, 7), 0,                 _R(VOPENID1, 7), 0,                 _R(VHHENID1, 7),  0                },

  /* 17 */ { 0, _R(TTLID2, 0), 0,                 _R(VOPENID2, 0), 0,                 _R(VHHENID2, 0),  0                },
  /* 18 */ { 0, _R(TTLID2, 1), 0,                 _R(VOPENID2, 1), 0,                 _R(VHHENID2, 1),  0                },
  /* 19 */ { 0, _R(TTLID2, 2), _R(OTHERENID1, 0), _R(VOPENID2, 2), 0,                 _R(VHHENID2, 2),  0                },
  /* 20 */ { 0, _R(TTLID2, 3), _R(OTHERENID1, 1), _R(VOPENID2, 3), 0,                 _R(VHHENID2, 3),  0                },
  /* 21 */ { 0, _R(TTLID2, 4), _R(OTHERENID1, 2), _R(VOPENID2, 4), 0,                 _R(VHHENID2, 4),  0                },
  /* 22 */ { 0, _R(TTLID2, 5), 0,                 _R(VOPENID2, 5), 0,                 _R(VHHENID2, 5),  0                },
  /* 23 */ { 0, _R(TTLID2, 6), 0,                 _R(VOPENID2, 6), 0,                 _R(VHHENID2, 6),  0                },
  /* 24 */ { 0, _R(TTLID2, 7), _R(VCCENID1, 7),   _R(VOPENID2, 7), 0,                 _R(VHHENID2, 7),  0                },

  /* 25 */ { 0, _R(TTLID3, 0), _R(VCCENID1, 6),   _R(VOPENID3, 0), _R(OTHERENID0, 2), _R(VHHENID3, 0),  0                },
  /* 26 */ { 0, _R(TTLID3, 1), _R(VCCENID1, 1),   _R(VOPENID3, 1), 0,                 _R(VHHENID3, 1),  0                },
  /* 27 */ { 0, _R(TTLID3, 2), _R(VCCENID1, 0),   _R(VOPENID3, 2), 0,                 _R(VHHENID3, 2),  0                },
  /* 28 */ { 0, _R(TTLID3, 3), _R(VCCENID0, 7),   _R(VOPENID3, 3), 0,                 _R(VHHENID3, 3),  _R(VHHENCID0, 7) },
  /* 29 */ { 0, _R(TTLID3, 4), _R(VCCENID0, 6),   _R(VOPENID3, 4), 0,                 _R(VHHENID3, 4),  _R(VHHENCID0, 6) },
  /* 30 */ { 0, _R(TTLID3, 5), _R(VCCENID0, 5),   _R(VOPENID3, 5), 0,                 _R(VHHENID3, 5),  _R(VHHENCID0, 5) },
  /* 31 */ { 0, _R(TTLID3, 6), _R(VCCENID0, 4),   _R(VOPENID3, 6), 0,                 _R(VHHENID3, 6),  _R(VHHENCID0, 4) },
  /* 32 */ { 0, _R(TTLID3, 7), _R(VCCENID0, 3),   _R(VOPENID3, 7), 0,                 _R(VHHENID3, 7),  _R(VHHENCID0, 3) },

  /* 33 */ { 0, _R(TTLID4, 0), 0,                 0,               0,                 0,                0                },
  /* 34 */ { 0, _R(TTLID4, 1), _R(VCCENID0, 2),   0,               0,                 0,                0                },
  /* 35 */ { 0, _R(TTLID4, 2), 0,                 0,               0,                 0,                0                },
  /* 36 */ { 0, _R(TTLID4, 3), _R(VCCENID0, 1),   _R(VOPENID4, 3), 0,                 0,                0                },
  /* 37 */ { 0, _R(TTLID4, 4), 0,                 0,               0,                 0,                0                },
  /* 38 */ { 0, _R(TTLID4, 5), 0,                 0,               0,                 0,                0                },
  /* 39 */ { 0, _R(TTLID4, 6), 0,                 0,               0,                 0,                0                },
  /* 40 */ { 0, _R(TTLID4, 7), _R(VCCENID0, 0),   0,               0,                 0,                0                }
};

static const uint8_t gnd_lut[PIN_COUNT] = {
  /* pin 1 ... */ 0x40,    0,    0,    0,    0,    0,    0,    0,
  /* pin 9 ... */    0,    0, 0x01,    0,    0,    0,    0,    0,
  /* pin 17... */    0,    0,    0, 0x00, 0x41,    0,    0,    0,
  /* pin 25... */    0,    0,    0, 0x80, 0x81, 0xc0,    0,    0,
  /* pin 33... */    0,    0,    0,    0,    0,    0,    0, 0xc1,
};

static const uint8_t gnd_lut2[8] = {
  20, 11, 1, 21, 28, 29, 30, 40
};

/* set pin as input, disable all drivers but PIN_M_GND */
static void all3_set_pin_input(uint16_t pin)
{
  short mode;

  /* special case for clock and 100nf capacitor */
  switch (pin)
    {
    case 2:                     /* clock */
    case 3:
      all3_reg_or(REG_OTHERENID0, 0, 1 << 3);
      if (pin_mode[2-1] == PIN_M_CLK ||
          pin_mode[3-1] == PIN_M_CLK)
        {
          pin_mode[2-1] = PIN_M_INPUT;
          pin_mode[3-1] = PIN_M_INPUT;
        }
      break;

    case 18:                    /* clock */
    case 19:
      all3_reg_or(REG_OTHERENID0, 0, 1 << 4);
      if (pin_mode[18-1] == PIN_M_CLK ||
          pin_mode[19-1] == PIN_M_CLK)
        {
          pin_mode[18-1] = PIN_M_INPUT;
          pin_mode[19-1] = PIN_M_INPUT;
        }
      break;

    case 30:                    /* 100nF capacitor & Alt GND */
      all3_reg_or(REG_OTHERENID1, 0, 1 << 5 | 1 << 6);
      break;

    case 32:                    /* 100nF capacitor */
      all3_reg_or(REG_OTHERENID1, 0, 1 << 7);
      break;
    }

  /* generic case, rely on table pin_reg_ops to disable any mode */
  for (mode = PIN_M_VHH2; mode >= PIN_M_INPUT; mode--)
    {
      uint16_t op = pin_reg_ops[pin - 1][mode];

      if (op & 0x80)    /* is mode implemented for this pin ? */
        {
          uint8_t id = op >> 8;
          uint8_t bit = 1 << (op & 7);
          uint8_t x = all3_reg_get(id);

          switch (mode)
            {
            case PIN_M_INPUT:
              x |= bit;
              break;

            case PIN_M_VOP:
            case PIN_M_VOP25:
              /* update count of VOP enabled pins */
              if (x & bit)
                vop_count--;
              x &= ~bit;
              break;

            case PIN_M_VHH:
            case PIN_M_VHH2:
              /* update count of VHH enabled pins */
              if (x & bit)
                vhh_count--;
              x &= ~bit;
              break;

            default:
              x &= ~bit;
              break;
            }

          all3_reg_set(id, x);
        }
    }

  if (pin_mode[pin - 1] != PIN_M_GND)
    pin_mode[pin - 1] = PIN_M_INPUT;
}

static void all3_set_pin_mode(uint16_t pin, uint8_t mode)
{
  uint16_t cap = pin_capability[pin - 1];
  uint8_t cur_mode = pin_mode[pin - 1];
  uint8_t pin_reg = (pin - 1) / 8;
  uint8_t pin_bit = (pin - 1) % 8;
  uint8_t pin_mask = 1 << pin_bit;

  if (!((cap >> mode) & 1))
    return;
  if (mode == cur_mode)
    return;

  /* not possbile to disable all GND, one has to be left enabled */
  if (cur_mode == PIN_M_GND)
    all3_set_pin_mode(pin == 20 ? 1 : 20, PIN_M_GND);

  /* disable current mode if not GND */
  all3_set_pin_input(pin);

  /* set new mode */
  switch (mode)
    {
    case PIN_M_GND: {
      /* find current GND */
      uint8_t x = all3_reg_get(REG_OTHERENID0) & 0xc1;
      x = gnd_lut2[(x >> 5) | (x & 1)];
      /* current GND is reverted back to input */
      pin_mode[x - 1] = PIN_M_INPUT;

      /* select new GND */
      all3_reg_or(REG_OTHERENID0, gnd_lut[pin - 1], 0xc1);
      break;
    }

    case PIN_M_INPUT:
      /* nothing to do */
      break;

    case PIN_M_TTLLOW:
      all3_reg_or(REG_TTLID0 + pin_reg, 0, pin_mask);
      break;

    case PIN_M_VHH:
    case PIN_M_VHH2:
      if (vhh_count >= 3 &&
        (gui_query_bar("Many VHH pins already enabled, are you sure? (y/n)") | 32) != 'y')
        return;
      vhh_count++;
      goto generic_enable;
    case PIN_M_VOP25:
    case PIN_M_VOP:
      if (vop_count >= 3 &&
        (gui_query_bar("Many VOP pins already enabled, are you sure? (y/n)") | 32) != 'y')
        return;
      vop_count++;
    generic_enable:
    case PIN_M_VCC: {
      uint16_t op = pin_reg_ops[pin - 1][mode];
      all3_reg_or(op >> 8, 0xff, 1 << (op & 7));
      break;
    }

    case PIN_M_CLK: {           /* clock pairs */
      uint8_t pin2 = pin ^ 1;
      all3_set_pin_input(pin2);
      switch (pin)
        {
        case 2:
        case 3:
          pin_mode[pin2 - 1] = mode;
          all3_reg_or(REG_OTHERENID0, 0xff, 1 << 3);
          break;

        case 18:
        case 19:
          pin_mode[pin2 - 1] = mode;
          all3_reg_or(REG_OTHERENID0, 0xff, 1 << 4);
          break;
        }
      break;
    }

    case PIN_M_100N:            /* 100nF capacitor */
      switch (pin)
        {
        case 30:
          all3_reg_or(REG_OTHERENID1, 0xff, 1 << 6);
          break;

        case 32:
          all3_reg_or(REG_OTHERENID1, 0xff, 1 << 7);
          break;
        }
      break;

    case PIN_M_GND2:            /* pin 30 only */
      all3_reg_or(REG_OTHERENID1, 0xff, 1 << 5);
      break;
    }

  pin_mode[pin - 1] = mode;
}

/* this guesses pin voltage depending on current pin mode and DAC values */
static int16_t all3_get_pin_voltage(short pin)
{
  uint8_t cur_mode = pin_mode[pin - 1];
  int16_t v;

  switch (cur_mode)
    {
    case PIN_M_TTLLOW:
    case PIN_M_GND:
    case PIN_M_GND2:
      return 0;

    case PIN_M_VCC:
      v = vcc_dac - 60;         /* diode drop */
      break;
    case PIN_M_VOP:
      v = vop_dac - 60;
      break;
    case PIN_M_VOP25:
      v = vop_dac - 60 - 250;
      break;
    case PIN_M_VHH:
    case PIN_M_VHH2:
      v = vhh_dac - 60;
      break;

    default:
      return -1;
    }

  return v < 0 ? 0 : v;
}

static void draw_status_bar(void)
{
  _settextcolor(0);
  _setbkcolor(_WHITE);
  _settextposition(25, 1);
  _outtext(" All-03 & Expro60 test software v1.1       \xb3 \x11\x1e\x1f\x10: Select pin \xb3 ?:Help \xb3 q:Quit ");
}

static char gui_query_bar(const char *message)
{
  uint16_t k, len = strlen(message);
  _settextcolor(16+15);
  _setbkcolor(_RED);
  _settextposition(25, 1);
  _outtext(" ");
  _outmem(message, len);
  _outmem("                                                                                ", 80 - len - 1);

  k = _bios_keybrd(_KEYBRD_READ);
  draw_status_bar();

  return k & 255;
}

/* graph lib bug workaround for char > 127, need to add 1 to
   _settextcolor argument to display non-Ascii characters with correct
   color. */
#define TEXTCOLOR_FIX 1

static void draw_dip40(short x, short y, short pin)
{
  uint16_t i;

  _setbkcolor(_BLACK);

  /* draw package top side */
  _settextposition(y, x + 10);
  _settextcolor(7 + TEXTCOLOR_FIX);
  _outtext("  \xda\xc4\xc4\xc4\xc4_\xc4\xc4\xc4\xc4\xbf  ");

  /* draw 2 rows of 20 pins */
  for (i = 1; i <= 20; i++)
    {
      char s[16];
      uint16_t p = i;
      _settextposition(y + p, x + 10);

      sprintf(s, " %02u    ", p);
      _setbkcolor(_BLACK);
      _settextcolor(p == pin ? 12 : 8);
      _outtext(s);

      p = 41 - i;
      sprintf(s, "     %2u ", p);
      _settextcolor(p == pin ? 12 : 8);
      _outtext(s);
    }

  /* draw package bottom side */
  _settextposition(y + 21, x + 10);
  _setbkcolor(_BLACK);
  _settextcolor(7 + TEXTCOLOR_FIX);
  _outtext("  \xc0\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xd9  ");
}

static void draw_dip40_status_mode(short x, short y,
                        uint16_t pin, uint16_t cur_pin)
{
  const char *t;
  char s[16];
  uint8_t mode = pin_mode[pin - 1];

  /* pin status string */
  if (mode == PIN_M_INPUT)
    t = all3_read_pin(pin) ? "[High]" : "[Low]";
  else
    t = pin_cap_desc[mode];

  _setbkcolor(_BLUE);
  sprintf(s, pin > 20 ? "%-9s" : "%9s", t);
  _outtext(s);

  /* possibly display expected voltage */
  if (pin == cur_pin)
    {
      int16_t v = all3_get_pin_voltage(pin);
      _setbkcolor(_BLACK);
      _settextposition(y, x);
      if (v >= 0)
        {
          v += 5;
          sprintf(s, pin > 20 ? "%2u.%uv" : "%u.%uv",
            v / 100, (v % 100) / 10);
          _outtext(s);
        }
      else
        {
          _outtext("     ");
        }
    }
}

static void draw_dip40_status(short x, short y, uint16_t pin)
{
  uint16_t i;

  _settextcolor(7);

  /* draw status for all 40 pins */
  for (i = 1; i <= 20; i++)
    {
      _settextposition(y + i, x);
      draw_dip40_status_mode(x + 14, y + i, i, pin);

      _settextposition(y + i, x + 26);
      draw_dip40_status_mode(x + 16, y + i, 41 - i, pin);
    }
}

static void draw_base_addr(short x, short y)
{
  char s[32];
  _settextcolor(0);
  _setbkcolor(_WHITE);
  _settextposition(y, x);
  sprintf(s, "  IO addr= %04X  ", base_addr);
  _outtext(s);
}

/* draw box with list of possible pin modes */
static void draw_pin_drive(short x, short y, short pin)
{
  uint16_t i;
  uint16_t cap = pin_capability[pin - 1];
  char s[32];

  _settextcolor(0);
  _setbkcolor(_WHITE);
  _settextposition(y, x);
  sprintf(s, "     Pin %-2u      ", pin);
  _outtext(s);

  _setbkcolor(_BLACK);

  for (i = 0; i < PIN_M_COUNT; i++)
    {
      sprintf(s, " %x: %-10s   ", i, pin_cap_desc[i]);

      if (pin_mode[pin - 1] == i)
        _settextcolor(12);
      else
        _settextcolor((cap >> i) & 1 ? 15 : 8);

      _settextposition(y + 1 + i, x);
      _outtext(s);
    }
}

static void draw_regs(short x, short y)
{
  uint8_t i;
  char s[32];

  _settextcolor(0);
  _setbkcolor(_WHITE);
  _settextposition(y++, x);
  _outtext("   Registers   ");

  _setbkcolor(_BLACK);
  _settextcolor(7);

  for (i = 0; i < 24; i++)
    {
      static const char *reg_name[24] = {
        "TTLID0",       "TTLID1",       "TTLID2",       "TTLID3",
        "TTLID4",       "VOPID",        "VHHID",        "VCCID",
        "VOPENID0",     "VOPENID1",     "VOPENID2",     "VOPENID3",
        "VOPENID4",     "VCCENID0",     "VCCENID1",     "OTHERENID0",
        "VHHENID0",     "VHHENID1",     "VHHENID2",     "VHHENID3",
        "VHHENID4",     "VHHENCID0",    "VHHENCID1",    "OTHERENID1"
      };

      switch (i + 0xe0)
        {
        case REG_VHHENCID1:
        case REG_VHHENID4:
        case REG_VHHENID0:
          continue;             /* not implemented in the hardware */
        }

      _settextposition(y++, x);
      sprintf(s, " %-10s %02X ", reg_name[i], reg_val[i]);
      _outtext(s);
    }
}

static void draw_dac_value(short x, short y)
{
  char s[32];

  _settextcolor(0);
  _setbkcolor(_WHITE);
  _settextposition(y, x);
  _outtext("   DAC & Clock   ");

  _setbkcolor(_BLACK);
  _settextcolor(7);
  _settextposition(y + 1, x);
  sprintf(s, " C/c: VCC=%2u.%02uv ", vcc_dac / 100, vcc_dac % 100);
  _outtext(s);

  _settextposition(y + 2, x);
  sprintf(s, " O/o: VOP=%2u.%02uv ", vop_dac / 100, vop_dac % 100);
  _outtext(s);

  _settextposition(y + 3, x);
  sprintf(s, " H/h: VHH=%2u.%02uv ", vhh_dac / 100, vhh_dac % 100);
  _outtext(s);

  _settextposition(y + 4, x);
  sprintf(s, " f: CLK=%sMhz ", all3_reg_get(REG_OTHERENID0) & (1 << 5)
          ? "4.432" : "2.216");
  _outtext(s);
}

/* draw dialog box borders */
static void draw_box(uint16_t x, uint16_t y,
  uint16_t width, uint16_t height, const char *title)
{
  int16_t i;

  _setbkcolor(_CYAN);
  _settextcolor(1 + TEXTCOLOR_FIX);

  if (width < 2 || height < 2)
    return;

  /* top border */
  _settextposition(y, x);
  _outtext("\xda");
  for (i = width - 2; i > 0; i -= 8)
    _outmem("\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4", i > 8 ? 8 : i);
  _outtext("\xbf");

  if (title)
    {
      int16_t tx = (width - strlen(title) - 2) / 2;
      if (tx > 1)
        {
          _settextposition(y, x + tx);
          _outmem("\xb4", 1);
          _settextcolor(8);
          _outtext(title);
          _settextcolor(1 + TEXTCOLOR_FIX);
          _outmem("\xc3", 1);
        }
    }

  /* left and right borders */
  for (i = 1; i < height - 1; i++)
    {
      _settextposition(y + i, x);
      _outtext("\xb3");

      _settextposition(y + i, x + width - 1);
      _outtext("\xb3");
    }

  /* bottom border */
  _settextposition(y + height - 1, x);
  _outtext("\xc0");
  for (i = width - 2; i > 0; i -= 8)
    _outmem("\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4", i > 8 ? 8 : i);
  _outtext("\xd9");
}

static const char *
draw_text_area(uint16_t x, uint16_t y, uint16_t width, uint16_t height,
               const char *text, uint16_t line)
{
  int16_t i = 0;

  if (width < 2 || height < 1)
    return 0;

  while (i < height)
    {
      const char *next = strchr(text, '\n');
      uint16_t len = next ? next - text : strlen(text);
      uint8_t skip_blank = 1;

      if (len > width)
        {
          /* line wrap */
          len = width;
          while (1)
            {
              if (len < 10 || len < width / 2)
                {
                  len = width;
                  skip_blank = 0;
                  break;        /* wrap at width */
                }
              if (text[--len] == ' ')
                break;  /* wrap at text blank */
            }
        }

      if (!line)
        {
          _settextposition(y + i, x);
          _outmem(text, len);
        }

      skip_blank &= !!text[len];
      text += len + skip_blank;

      if (!line)
        {
          /* pad end of line */
          int16_t j;
          for (j = width - len; j > 0; j -= 8)
            _outmem("        ", j > 8 ? 8 : j);
          i++;
        }
      else
        {
          line--;
        }
    }

  return text;
}

static void gui_text_window(uint16_t x, uint16_t y,
                        uint16_t width, uint16_t height,
                        const char *text, const char *title)
{
  uint16_t k, offset = 0;

  if (width < 5 || height < 3)
    return;

  draw_box(x, y, width, height, title);

  while (1)
    {
      const char *next;
      _settextcolor(0);
      next = draw_text_area(x + 1, y + 1, width - 2, height - 2, text, offset);

      k = _bios_keybrd(_KEYBRD_READ);
      switch (k >> 8)           /* key scan code */
        {
        case 0x50:              /* down */
          if (*next)
            offset++;
          break;
        case 0x51:              /* page down */
          if (*next)
            offset += height - 2;
          break;
        case 0x48:              /* up */
          offset -= !!offset;
          break;
        case 0x49:              /* page up */
          if (offset < height - 2)
            offset = 0;
          else
            offset -= height - 2;
          break;
        case 0x47:              /* home */
          offset = 0;
          break;
        default:
          return;
        }
    }
}

/* main GUI loop */
static void gui_pin_testing()
{
  uint16_t pin = 1;

 redraw_all:
  _setbkcolor(_BLUE);
  _settextcolor(15);
  _clearscreen(_GCLEARSCREEN);

  draw_status_bar();

  draw_base_addr(23, 23);
  draw_pin_drive(23, 2, pin);
  draw_dac_value(23, 16);
  draw_regs(4, 2);
  draw_dip40(43, 2, pin);
  draw_dip40_status(43, 2, pin);

  while (1)
    {
      uint16_t k = _bios_keybrd(_KEYBRD_READY);
      uint8_t sc = k >> 8;

      if (!k)
        {
          /* no key pressed yet */
          delay(100);
          goto redraw_input;
        }
      else
        {
          _bios_keybrd(_KEYBRD_READ);
        }

      k &= 255;
      switch (k)                /* key ascii */
        {
        case 'A':
        case 'a':
          k = '0' + 10;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
          k -= '0';
          /* pin mode change requested */
          if ((pin_capability[pin - 1] >> k) & 1)
            all3_set_pin_mode(pin, k);
          goto redraw_status;

          /* VCC DAC change requested */
        case 'c':
          if (vcc_dac > VCC_DAC_MIN)
            vcc_dac -= VCC_DAC_STEP;
          goto vcc_dac_update;
        case 'C':
          if (vcc_dac < VCC_DAC_MAX)
            vcc_dac += VCC_DAC_STEP;
        vcc_dac_update:
          all3_reg_set(REG_VCCID, (vcc_dac - VCC_DAC_MIN) / VCC_DAC_STEP);
          goto redraw_status;

          /* VOP DAC change requested */
        case 'o':
          if (vop_dac > VOP_DAC_LOW)
            vop_dac -= VOP_DAC_STEP;
          goto vop_dac_update;
        case 'O':
          if (vop_dac < VOP_DAC_MAX)
            vop_dac += VOP_DAC_STEP;
        vop_dac_update:
          all3_reg_set(REG_VOPID, (vop_dac - VOP_DAC_MIN) / VOP_DAC_STEP);
          goto redraw_status;

          /* VHH DAC change requested */
        case 'h':
          if (vhh_dac > VHH_DAC_LOW)
            vhh_dac -= VHH_DAC_STEP;
          goto vhh_dac_update;
        case 'H':
          if (vhh_dac < VHH_DAC_MAX)
            vhh_dac += VHH_DAC_STEP;
        vhh_dac_update:
          all3_reg_set(REG_VHHID, (vhh_dac - VHH_DAC_MIN) / VHH_DAC_STEP);
          goto redraw_status;

          /* clock frequency change requested */
        case 'f':
          all3_reg_or(REG_OTHERENID0, ~all3_reg_get(REG_OTHERENID0), 1 << 5);
          goto redraw_status;

          /* help requested */
        case '?':
          gui_text_window(5, 5, 72, 16,
            "This third party software allows testing the hardware of your "
            "Hi-Lo All-03, Sunshine Expro 60 or compatible programmer.\n\n"

            "First, remove any IC present in the programmer socket. Then, use "
            "the arrow keys to select the pin of the DIP socket you "
            "want to test. When in input mode, the logic level sensed by the "
            "programmer is displayed as [Low] or [High]. You can press a "
            "numeric key to change the pin driving mode. Modes that are not "
            "implemented in the hardware for a given pin are grayed out. You may "
            "then use a voltmeter to check that the actual voltage matches "
            "what is predicted by the software for that pin. The VCC, VOP and "
            "VHH internal voltages can be changed by pressing the C, c, O, o, "
            "H and h keys.\n\n"
            LICENCE_TEXT, "Help");
          goto redraw_all;

          /* quit to DOS requested */
        case 'Q':
        case 'q':
          if ((gui_query_bar("Quit to DOS? (y/n)") | 32) == 'y')
            return;
          break;

        default:
          break;
        }

      /* navigate through DIP package pins layout */
      switch (sc)               /* key scan code */
        {
        case 0x48:              /* up */
          if (pin > 20)
            pin += pin < 40;
          else
            pin -= pin > 1;
          goto redraw_pins;
        case 0x50:              /* down */
          if (pin > 20)
            pin -= pin > 21;
          else
            pin += pin < 20;
          goto redraw_pins;
        case 0x4b:              /* left */
          if (pin > 20)
            pin = 41 - pin;
          goto redraw_pins;
        case 0x4d:              /* right */
          if (pin < 21)
            pin = 41 - pin;
          goto redraw_pins;

        default:
          /* key not handled, ignore */
          continue;
        }

    redraw_pins:
      draw_dip40(43, 2, pin);
      draw_pin_drive(23, 2, pin);
      draw_dip40_status(43, 2, pin);
      continue;
    redraw_status:
      draw_regs(4, 2);
      draw_pin_drive(23, 2, pin);
      draw_dac_value(23, 16);
    redraw_input:
      draw_dip40_status(43, 2, pin);
      continue;
    }
}

static char all3_check_code()
{
  uint8_t i, x;
  static const uint8_t codes[16] = {
    /* as returned by Hi-Lo All-03 */
    14, 13, 11, 6, 12, 9, 2, 5, 10, 4, 8, 0, 1, 3, 7, 15
  };

  all3_reg_set(REG_VCCID, 0);
  all3_reg_set(REG_VCCID, 1);
  all3_reg_set(REG_VCCID, 0);
  all3_reg_set(REG_VCCID, 2);

  for (i = 0; i < 16; i++)
    {
      all3_reg_set(REG_VCCID, 3);
      all3_reg_set(REG_VCCID, 2);
      x = all3_reg_in(REG_CODEID) & 15;
      if (x != codes[i])
        break;
    }

  all3_reg_set(REG_VCCID, 0);

  return i == 16;
}

/* in case we have an unknown hardware */
static void all3_disp_code()
{
  uint8_t i;

  all3_reg_set(REG_VCCID, 0);
  all3_reg_set(REG_VCCID, 1);
  all3_reg_set(REG_VCCID, 0);
  all3_reg_set(REG_VCCID, 2);

  for (i = 0; i < 16; i++)
    {
      all3_reg_set(REG_VCCID, 3);
      all3_reg_set(REG_VCCID, 2);
      printf("(%i:%x)", i, all3_reg_in(REG_CODEID) & 15);
    }

  printf("\n");
  all3_reg_set(REG_VCCID, 0);
  _bios_keybrd(_KEYBRD_READ);
}

static void all3_regs_default()
{
  unsigned short x;

  for (x = 1; x <= PIN_COUNT; x++)
    pin_mode[x - 1] = PIN_M_INPUT;
  pin_mode[20 - 1] = PIN_M_GND;

  /* all pins as inputs */
  for (x = REG_TTLID0; x <= REG_TTLID4; x++)
    all3_reg_set(x, 0xff);
  /* disable other drivers and set DACs to min value. */
  for (x = REG_VOPID; x <= REG_OTHERENID1; x++)
    all3_reg_set(x, 0);
}

int main(int argc, char **argv)
{
  int16_t c;
  uint8_t no_hw_check = 0;

  while ((c = getopt(argc, argv, "dfa:")) > 0)
    switch (c)
      {
      case 'a':
        base_addr = strtoul(optarg, NULL, 16);
        break;
      case 'd':
        all3_disp_code();
        break;
      case 'f':
        no_hw_check = 1;
        break;
      default:
        printf("invalid option: %c\n", c);
        return 1;
      }

  if ((base_addr & 0xff0f) != 0x200)
    {
      printf("invalid IO base address: %04X\n", base_addr);
      return 1;
    }

  if (!no_hw_check && !all3_check_code())
    {
      printf("Programmer device not detected at %04X.\r\n"
             "Set address using -a XXXX or force with -f.\n", base_addr);
      return 1;
    }

  all3_regs_default();
  all3_reg_set(REG_VOPID, (vop_dac - VOP_DAC_MIN) / VOP_DAC_STEP);
  all3_reg_set(REG_VHHID, (vhh_dac - VHH_DAC_MIN) / VHH_DAC_STEP);

  _setvideomode(_TEXTC80);
  _displaycursor(_GCURSOROFF);
  _wrapon(_GWRAPOFF);

  gui_pin_testing();

  all3_regs_default();

  _setvideomode(_DEFAULTMODE);
  _displaycursor(_GCURSORON);
  _clearscreen(_GCLEARSCREEN);

  return 0;
}
